/**********************************************/
/* mpi_integral.c:                            */
/* Copyright (C) 2003 Tomonori Kouya          */
/*                                            */
/* Version 0.0: 2003.08/21                    */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/
#include <stdio.h>
#include "bnc.h"
#ifndef __MPI_GMP_H
	#include "mpi_gmp.h"
#endif

/* double: Trapezoidal rule */
void _mpi_dtrapezoidal_fs(double *ptr_ret, double x_start, double x_end, double (*func)(double x), long int num_div, MPI_Comm comm)
{
	static double x, h, local_ret;
	static long int i;

	int myrank, num_procs;

	MPI_Comm_size(comm, &num_procs);
	MPI_Comm_rank(comm, &myrank);

	/* set stepsize */
	h = (x_end - x_start) / num_div;

	/* ret := h * (f(a) + f(b))/2 */
	local_ret = 0.0;
	if(myrank == 0)
	{
		*ptr_ret = h * (func(x_start) + func(x_end)) / 2;
	}

	/* ret += sum^{n-1}_{i = 1} f(x_i) */
	for(i = myrank + 1; i < num_div; i += num_procs)
	{
		x = x_start + h * i;
		local_ret += func(x);
	}

	/* reduce */
	local_ret *= h;
//	printf("h = %e\n", h);
//	printf("local_ret: %e\n", local_ret);
	MPI_Reduce(&local_ret, ptr_ret, 1, MPI_DOUBLE, MPI_SUM, 0, comm);

}

#ifdef USE_GMP
/* Trapezoidal rule */
void _mpi_mpf_trapezoidal_fs(mpf_t ret, mpf_t x_start, mpf_t x_end, void (*func)(mpf_t, mpf_t), long int num_div, MPI_Comm comm, MPI_Datatype mpi_mpf_type)
{
	mpf_t x, h, tmp, local_ret;
	long int i;

	int myrank, num_procs;
	void *local_ret_buf, *ret_buf;

	MPI_Comm_size(comm, &num_procs);
	MPI_Comm_rank(comm, &myrank);

	mpf_init2(x, mpf_get_prec(x_start));
	mpf_init2(h, mpf_get_prec(x_start));
	mpf_init2(tmp, mpf_get_prec(ret));
	mpf_init2(local_ret, mpf_get_prec(ret));

	/* set Stepsize */
	mpf_sub(h, x_end, x_start); mpf_div_ui(h, h, (unsigned long)num_div);

	/* ret := h * (f(a) + f(b)) / 2 */
	func(ret, x_start);
	func(tmp, x_end);
	mpf_add(ret, ret, tmp); mpf_div_ui(ret, ret, 2UL);

	/* set starting and ending values */
	mpf_add(x, x_start, h);

	/* ret += sum^{n-1}_{i=1} f(x_i) */
	mpf_set_ui(local_ret, 0UL);
	for(i = myrank + 1; i < num_div; i += num_procs)
	{
		/* x = x_start + h * i */
		mpf_mul_ui(tmp, h, (unsigned long)i);
		mpf_add(x, x_start, tmp);
		func(tmp, x);
		mpf_add(local_ret, local_ret, tmp);
	}
//	mpf_mul(local_ret, local_ret, h);

//	printf("local_ret = "); mpf_out_str(stdout, 10, 0, local_ret); printf("\n");

	local_ret_buf = allocbuf_mpf(mpf_get_prec(local_ret), 1);
	ret_buf = allocbuf_mpf(mpf_get_prec(ret), 1);

	pack_mpf(local_ret, 1, local_ret_buf);
	pack_mpf(ret, 1, ret_buf);
	MPI_Reduce(local_ret_buf, ret_buf, 1, mpi_mpf_type, MPI_MPF_SUM, 0, comm);
	unpack_mpf(ret_buf, ret, 1);

	/* ret := h * { (f(a) + f(b)) / 2 + sum^{n-1}_{i=1} f(x_i) } */
	if(myrank == 0)
		mpf_mul(ret, ret, h);

	freebuf_mpf(local_ret_buf);
	freebuf_mpf(ret_buf);

	mpf_clear(x);
	mpf_clear(h);
	mpf_clear(local_ret);
	mpf_clear(tmp);
}
#endif
